"""
    This module contains a **constants** and **default** parameters. These parameters can be accessed 
    at any time by whatever of the modules. 

    Example:
        To use the ``Constants`` import the module as::

            >>> import symupy.utils.constants as ct
            >>> ct.BUFFER_STRING # access the buffer size 


    ============================  =================================
     **Variable**                 **Description**
    ----------------------------  ---------------------------------
    ``BUFFER_STRING``              Buffer size
    ``FIELD_DATA``                 Vehicle trajectory data
    ``FIELD_FORMAT``               Trajectory data types
    ``HOUR_FORMAT``                Time format
    ``FIELD_FORMATAGG``            Format aggretations
    ``DCT_SIMULATION_INFO```       XML Simulation information
    ``DCT_EXPORT_INFO``            XML Export information
    ``DCT_TRAFIC_INFO``            XML Traffic information
    ``DCT_NETWORK_INFO``           XML Network information
    ``DCT_SCENARIO_INFO``          XML Scenario information
    ``TP_VEHTYPES``                Vehicle type information
    ``TP_ACCEL``                   Vehicle acceleration boundaries
    ============================  =================================

"""

from datetime import date, datetime, timedelta
from numpy import array, float64, int32

# *****************************************************************************
# CONNECTOR
# *****************************************************************************
# Buffer string size
BUFFER_STRING = 1000000


# *****************************************************************************
# DATA SYMUVIA
# *****************************************************************************
HOUR_FORMAT = "%H:%M:%S"  # Format time from xml file


# *****************************************************************************
# DATA SYMUVIA
# *****************************************************************************
RADIOUS_ANT = 500  # meters. Default distance scope antenna before

FIELD_DATA = {
    "@abs": "abscisa",
    "@acc": "acceleration",
    "@dst": "distance",
    "@id": "vehid",
    "@ord": "ordinate",
    "@tron": "link",
    "@type": "vehtype",
    "@vit": "speed",
    "@voie": "lane",
    "@z": "elevation",
}

FIELD_FORMAT = {
    "@abs": float,
    "@acc": float,
    "@dst": float,
    "@id": int,
    "@ord": float,
    "@tron": str,
    "@type": str,
    "@vit": float,
    "@voie": int,
    "@z": float,
}

FLOAT_SELECT = float64
INT_SELECT = int32

FIELD_FORMATAGG = {
    "abscisa": (array, FLOAT_SELECT),
    "acceleration": (array, FLOAT_SELECT),
    "distance": (array, FLOAT_SELECT),
    "vehid": (array, INT_SELECT),
    "ordinate": (array, FLOAT_SELECT),
    "link": (list, None),
    "vehtype": (list, None),
    "speed": (array, FLOAT_SELECT),
    "lane": (array, INT_SELECT),
    "elevation": (array, FLOAT_SELECT),
}


# *****************************************************************************
# XML Data
# *****************************************************************************

# DATE/TIME INFORMATION
DELTA_TIME = timedelta(minutes=1)
TIME_STEP = timedelta(seconds=1).total_seconds()
today = date.today().strftime("%Y-%m-%d")
st_time = datetime.now()
ed_time = st_time + DELTA_TIME
st_time_str = st_time.strftime("%H:%M:%S")
ed_time_str = ed_time.strftime("%H:%M:%S")

# SIMULATION INFORMATION
DCT_SIMULATION_INFO = {
    "id": "simID",
    "pasdetemps": f"TIME_STEP",
    "debut": f"st_time_str",
    "fin": f"ed_time_str",
    "loipoursuite": "exacte",
    "comportementflux": "iti",
    "date": f"today",
    "titre": "default_simulation",
    "proc_deceleration": "false",
    "seed": "1",
}

# DATA EXPORT INFORMATION
DCT_EXPORT_INFO = {
    "trace_route": "false",
    "trajectoires": "true",
    "debug": "false",
    "debug_matrice_OD": "false",
    "debug_SAS": "false",
    "csv": "true",
}

# TAFFIC INFORMATION
DCT_TRAFIC_INFO = {
    "id": "trafID",
    "accbornee": "true",
    "coeffrelax": "4",
    "chgtvoie_ghost": "false",
}

# NETWORK INFORMATION
DCT_NETWORK_INFO = {"id": "resID"}

# SCENARIO INFORMATION
DCT_SCENARIO_INFO = {
    "id": "defaultScenario",
    "simulation_id": DCT_SIMULATION_INFO.get("id"),
    "trafic_id": DCT_TRAFIC_INFO.get("id"),
    "reseau_id": DCT_NETWORK_INFO.get("id"),
    "dirout": "data",
    "prefout": "simout_ring_data",
}


TP_VEHTYPES = (
    {"id": "HDV", "w": "-5", "kx": "0.12", "vx": "25"},
    {"id": "CAV", "w": "-5", "kx": "0.12", "vx": "25"},
)

TP_ACCEL = (
    {"ax": "1.5", "vit_sup": "5.8"},
    {"ax": "1", "vit_sup": "8"},
    {"ax": "0.5", "vit_sup": "infini"},
)

# *****************************************************************************
# DATA CONTROL
# *****************************************************************************
BUFFER_CONTROL = 10  # Amount of control samples stored in memory

# *****************************************************************************
# DATA VEHICLE DYNAMICS
# *****************************************************************************
ENGINE_CONSTANT = 0.2



"""
    This module contains a ``Configurator`` object. The configurator is an object that stores *parameters* that can be relevant to make the evolution of a simulation. The objective is to introduce flexibility when configuring the the simulator platform and the runtime execution possibilities offered by exposed functions from the c library of SymuVia
"""

# ============================================================================
# INTERNAL IMPORTS
# ============================================================================

import symupy.utils.constants as CT

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================


class Configurator(object):
    """ Configurator class for containing specific simulator parameters
    
        Example:
            To use the ``Simulator`` declare in a string the ``path`` to the simulator ::

                >>> path = "path/to/simluator.so"
                >>> simulator = Configurator(libraryPath = path) 


        :return: Configurator object with simulation parameters
        :rtype: Configurator
    """

    def __init__(
        self,
        bufferSize: int = CT.BUFFER_STRING,
        writeXML: bool = True,
        traceFlow: bool = False,
        libraryPath: str = "",
        totalSteps: int = 0,
        stepLaunchMode: str = "lite",
    ) -> None:
        """ Configurator class for containing specific simulator parameter
        
            :param bufferSize: Provide an integer for buffer, defaults to CT.BUFFER_STRING
            :type bufferSize: int, optional
            :param writeXML: Flag to write XML file, defaults to True
            :type writeXML: bool, optional
            :param traceFlow: Flag to trace Flow / Traces, defaults to False
            :type traceFlow: bool, optional
            :param libraryPath: Stores the path of a traffic simulator, defaults to ""
            :type libraryPath: str, optional
            :param totalSteps: total number of simulation steps, defaults to 0
            :type totalSteps: int, optional
            :param stepLaunchMode: lite / full, defaults to "lite"
            :type stepLaunchMode: str, optional
            :return: Configurator object with simulation parameters
            :rtype: Configurator
        """
        self.bufferSize = bufferSize
        self.writeXML = writeXML
        self.traceFlow = traceFlow
        self.libraryPath = libraryPath
        self.totalSteps = totalSteps
        self.stepLaunchMode = stepLaunchMode

    def __repr__(self):
        data_dct = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"{self.__class__.__name__}({data_dct})"

    def __str__(self):
        data_dct = "Configuration status:\n " + "\n ".join(f"{k}:  {v}" for k, v in self.__dict__.items())
        return f"{data_dct}"



""" 
    Road Networks - Submodule

    This submodule contains the networks that can be created at infrastructure level. Networks can be of different types and at different layers. E.g. Road network where cars drive is physical one. 

    LICIT-LAB 
    Author: Andres Ladino
"""
import typing

# Module imports
from symupy.utils import constants as ct

import networkx as nx


class RoadNetwork(object):
    pass


class RoadSideUnit(object):

    rsuid = 0

    def __init__(self, link: str, position: float, radious: int = ct.RADIOUS_ANT):
        self.__class__.rsuid += 1
        self._link = link
        self._pos = position
        self._radious = radious

    def get_abs_pos(self):
        raise NotImplementedError



""" 
    This module defines the basic states required to execute and launch a simulation. 

    The states are defined as: 

    * **Compliance**: This state is defined to check availability of files and candidates.
    * **Connect**: This state is defined to process 
    * **Initialize**: This state perform initialization tasks. In example, loading the file scenarios into the simulator. Declaring initial conditions for the platoon, etc. 
    * **Preroutine**: Tasks to be done before the querying information from the simulator
    * **Query**: Tasks of parsing data and querying information from the simulator
    * **Control**: Perform decision tasks for the platoon 
    * **Push**: Push updated information to the simulator for platoon vehicles 
    * **Postroutine**: Performs tasks after the information has been pushed. 
"""

from .base import State


# Start of our states
class Compliance(State):
    """
    The state which declares an status to check file compliance .
    """

    def on_event(self, event: str, configurator) -> None:
        if event == "connect":
            return Connect()

        return self


class Connect(State):
    """
    The state which declares the creation of a connection with the simulator
    """

    def on_event(self, event: str, configurator):
        if event == "initialize":
            return Initialize()

        return self


class Initialize(State):
    """
    The state which initializes values for the scenario simulation
    """

    def on_event(self, event: str, configurator):
        if event == "preroutine":
            return PreRoutine()

        return self


class PreRoutine(State):
    """
    The state which performs task previous to the interaction with the simulator
    """

    def on_event(self, event: str, configurator):
        if event == "query":
            return Query()

        return self


class Query(State):
    """
    The state which retrieves information from the simulator
    """

    def on_event(self, event: str, configurator):
        if event == "control":
            return Control()

        return self


class Control(State):
    """
    The state which computes the control decision  
    """

    def on_event(self, event: str, configurator):
        if event == "push":
            return Push()

        return self


class Push(State):
    """
    The state which pushes data back to the simulator
    """

    def on_event(self, event: str, configurator):
        if event == "postroutine":
            return PostRoutine()

        return self


class PostRoutine(State):
    """
    The state which logs information or compute step indicators 
    """

    def on_event(self, event: str, configurator):
        if event == "preroutine":
            return PreRoutine()
        elif event == "terminate":
            return Terminate()

        return self


class Terminate(State):
    """
    The state which declares the end of a simulation
    """

    def on_event(self, event: str, configurator):
        return self


# End of our states.



"""
    This module describes classes and objects to perform a runtime of a single scenario
"""

# ============================================================================
# STANDARD  IMPORTS
# ============================================================================

from itertools import chain
import click

# ============================================================================
# INTERNAL IMPORTS
# ============================================================================

from .states import Compliance, Connect, Initialize, PreRoutine, Query, Control, Push, PostRoutine, Terminate
from symupy.utils import Configurator

# ============================================================================
# CLASS AND DEFINITIONS
# ============================================================================

start_seq = ["compliance", "connect", "initialize"]
runtime_seq = ["preroutine", "query", "control", "push", "postroutine"]
end_seq = [
    "terminate",
]


class RuntimeDevice(object):
    """ This class defines the runtime device describing a series of cyclic states required to be run 

    :param object: Runtime Device for controlling states of the simulation runtime
    :type object: Class with internal logic
    """

    def __init__(self, configurator: Configurator = Configurator()):
        self.state = Compliance()  # Initial state
        self.configurator = configurator
        self.cycles = configurator.total_steps

    def __enter__(self) -> None:
        """ Implementation of the state machine         
        """
        full_seq = chain(start_seq, self.cycles * runtime_seq, end_seq)

        currentCycle = 0
        for event in full_seq:
            self.on_event(event)
            if isinstance(self.state, PostRoutine):  # Step counted on PreRoutine
                currentCycle = currentCycle + 1
                click.echo(click.style(f"Step: {currentCycle}", fg="cyan", bold=True))
            if isinstance(self.state, Terminate):
                break

        self.on_event(event)  # Run Terminate sequence

        return self

    def __exit__(self, type, value, traceback) -> bool:
        return False

    def on_event(self, event: str):
        """ Action to consider on event:

        * compliance
        * connect
        * initialize
        * preroutine
        * query
        * control
        * push
        * postroutine
        * terminate
        
        :param event: 
        :type event: str 
        :param configurator:
        :type configurator: Configurator
        """
        self.state = self.state.on_event(event, self.configurator)
